@@ -3,13 +3,12 @@ import abc
 import numpy as np
 import scipy as sp
 
-from pygradflow.implicit_func import ImplicitFunc
 from pygradflow.iterate import Iterate
 from pygradflow.log import logger as lgg
 from pygradflow.params import NewtonType, Params
 from pygradflow.problem import Problem
-from pygradflow.step import step_solver
-from pygradflow.step.step_solver import StepResult, StepSolver
+from pygradflow.step.solver import step_solver
+from pygradflow.step.solver.step_solver import StepResult, StepSolver
 
 logger = lgg.getChild("newton")
 
@@ -22,7 +21,6 @@ class NewtonMethod(abc.ABC):
         self.orig_iterate = orig_iterate
         self.dt = dt
         self.rho = rho
-        self.func = ImplicitFunc(problem, orig_iterate, dt)
 
     @property
     def params(self) -> Params:
@@ -48,6 +46,7 @@ class SimplifiedNewtonMethod(NewtonMethod):
         step_solver: StepSolver,
     ) -> None:
         super().__init__(problem, orig_iterate, dt, rho)
+        self.func = step_solver.func
 
         self.step_solver = step_solver
         p = self.func.projection_initial(orig_iterate, rho)
@@ -76,6 +75,7 @@ class FullNewtonMethod(NewtonMethod):
         step_solver: StepSolver,
     ) -> None:
         super().__init__(problem, orig_iterate, dt, rho)
+        self.func = step_solver.func
         self.step_solver = step_solver
 
     def step(self, iterate: Iterate) -> StepResult:
@@ -97,6 +97,7 @@ class FixedActiveSetNewtonMethod(NewtonMethod):
 
     def __init__(self, problem, active_set, orig_iterate, dt, rho):
         super().__init__(problem, orig_iterate, dt, rho)
+        self.func = step_solver.func
 
         assert active_set.dtype == bool
         assert active_set.shape == problem.var_lb.shape
@@ -193,6 +194,7 @@ class ActiveSetNewtonMethod(NewtonMethod):
         step_solver: StepSolver,
     ) -> None:
         super().__init__(problem, orig_iterate, dt, rho)
+        self.func = step_solver.func
 
         self.step_solver = step_solver
         self.step_solver.update_derivs(orig_iterate)
@@ -206,6 +208,89 @@ class ActiveSetNewtonMethod(NewtonMethod):
         return self.step_solver.solve(iterate)
 
 
+class GlobalizedNewtonMethod(NewtonMethod):
+    """
+    Globalized Newton method with Armijo line search. Globalization
+    is based on the underlying function of the step solver.
+    """
+
+    def __init__(
+        self,
+        problem: Problem,
+        orig_iterate: Iterate,
+        dt: float,
+        rho: float,
+        step_solver: StepSolver,
+    ) -> None:
+        super().__init__(problem, orig_iterate, dt, rho)
+        self.func = step_solver.func
+        self.step_solver = step_solver
+
+    def _set_iterate(self, iterate):
+        self.step_solver.update_derivs(iterate)
+        p = self.func.projection_initial(iterate, self.rho)
+        active_set = self.func.compute_active_set(p)
+        self.step_solver.update_active_set(active_set)
+
+    def step(self, iterate):
+        problem = self.problem
+        params = iterate.params
+
+        self._set_iterate(iterate)
+
+        step_result = self.step_solver.solve(self.orig_iterate)
+
+        # Armijo line search:
+        alpha = 1.0
+
+        func_value = self.func.value_at(iterate, self.rho)
+        res_value = 0.5 * np.dot(func_value, func_value)
+
+        # TODO: Create a method in the step function
+        # to compute a forward product instead to make
+        # everything more efficient
+        func_deriv = self.func.deriv_at(iterate, self.rho)
+        func_grad = func_deriv.T @ func_value
+
+        func_grad_x = func_grad[: problem.num_vars]
+        func_grad_y = func_grad[problem.num_vars :]
+
+        dx = step_result.dx
+        dy = step_result.dy
+
+        inner_product = np.dot(func_grad_x, dx) + np.dot(func_grad_y, dy)
+
+        max_it = 30
+
+        for it in range(max_it):
+            next_iterate = Iterate(
+                problem, params, iterate.x - dx, iterate.y - dy, iterate.eval
+            )
+
+            next_func_value = self.func.value_at(next_iterate, self.rho)
+            next_res_value = 0.5 * np.dot(next_func_value, next_func_value)
+
+            if np.isclose(next_res_value, 0.0):
+                break
+
+            if next_res_value <= res_value + (1e-4 * alpha * inner_product):
+                break
+
+            alpha *= 0.5
+            dx = alpha * step_result.dx
+            dy = alpha * step_result.dy
+
+        else:
+            raise Exception("Line search failed to converge")
+
+        logger.debug("Line search converged in %d iterations", it + 1)
+
+        next_x = iterate.x + dx
+        active_set = self.func.compute_active_set(next_x)
+
+        return StepResult(self.orig_iterate, dx, dy, active_set, rcond=None)
+
+
 def newton_method(
     problem: Problem, params: Params, iterate: Iterate, dt: float, rho: float
 ) -> NewtonMethod:
@@ -220,3 +305,6 @@ def newton_method(
         return FullNewtonMethod(problem, iterate, dt, rho, solver)
     elif params.newton_type == NewtonType.ActiveSet:
         return ActiveSetNewtonMethod(problem, iterate, dt, rho, solver)
+    else:
+        assert params.newton_type == NewtonType.Globalized
+        return GlobalizedNewtonMethod(problem, iterate, dt, rho, solver)
