@@ -1,9 +1,13 @@
 import dataclasses
 import enum
+import typing
 from dataclasses import dataclass
 from enum import Enum, Flag, auto
 from typing import Any, Callable, Optional
 
+if typing.TYPE_CHECKING:
+    from .scale import Scaling
+
 import numpy as np
 
 
@@ -30,6 +34,11 @@ class NewtonType(Enum):
     Requires refactorizations but no reevaluations of the derivatives
     """
 
+    Globalized = auto()
+    """
+    Globalizes the Newton method by using an Armijo line search
+    """
+
 
 class StepSolverType(Enum):
     """
@@ -140,6 +149,16 @@ class ScalingType(Enum):
     Scale based on gradient and equilibration of constraint Jacobian
     """
 
+    KKT = auto()
+    """
+    Compute scales based on the equilibration of the KKT matrix
+    """
+
+    Nominal = auto()
+    """
+    Scaled based on values of the variable and constraint values
+    """
+
     Custom = auto()
     """
     User-provided custom scaling
@@ -191,7 +210,7 @@ class Params:
     precision: Precision = Precision.Double
 
     scaling_type: ScalingType = ScalingType.NoScaling
-    scaling: object = None
+    scaling: Optional["Scaling"] = None
 
     validate_input: bool = True
 
@@ -204,6 +223,7 @@ class Params:
     obj_lower_limit: float = -1e10
 
     report_rcond: bool = False
+    collect_path: bool = False
 
     def __post_init__(self):
         # Convert enum strings to enum values
