@@ -1,18 +1,16 @@
-import time
-from enum import Enum, auto
-from typing import Optional, cast
+from typing import List, Optional
 
 import numpy as np
 
-from pygradflow.display import Format, problem_display
-from pygradflow.eval import Evaluator, SimpleEvaluator, ValidatingEvaluator
+from pygradflow.callbacks import Callbacks, CallbackType
+from pygradflow.display import Format, StateData, print_problem_stats, solver_display
 from pygradflow.iterate import Iterate
 from pygradflow.log import logger
-from pygradflow.newton import newton_method
-from pygradflow.params import Params, ScalingType
+from pygradflow.params import Params
 from pygradflow.penalty import penalty_strategy
 from pygradflow.problem import Problem
-from pygradflow.scale import Scaling
+from pygradflow.result import SolverResult
+from pygradflow.status import SolverStatus
 from pygradflow.step.step_control import (
     StepController,
     StepControlResult,
@@ -21,132 +19,6 @@ from pygradflow.step.step_control import (
 from pygradflow.timer import Timer
 from pygradflow.transform import Transformation
 
-
-class SolverStatus(Enum):
-    Optimal = auto()
-    """
-    The algorithm has converged to a solution satisfying
-    the optimality conditions according to given tolerances
-    """
-
-    IterationLimit = auto()
-    """
-    Reached the iteration limit precribed by the algorithmic
-    parameters
-    """
-
-    TimeLimit = auto()
-    """
-    Reached the time limit precribed by the algorithmic
-    parameters
-    """
-
-    Unbounded = auto()
-    """
-    Problem appearst unbounded (found feasible point with extremely
-    small objective value)
-    """
-
-    LocallyInfeasible = auto()
-    """
-    Local infeasibility detected (found infeasible point being
-    a local minimum with respect to constraint violation)
-    """
-
-    @staticmethod
-    def short_name(status):
-        return {
-            SolverStatus.Optimal: "optimal",
-            SolverStatus.IterationLimit: "iteration_limit",
-            SolverStatus.TimeLimit: "time_limit",
-            SolverStatus.Unbounded: "unbounded",
-            SolverStatus.LocallyInfeasible: "infeasible",
-        }[status]
-
-    @staticmethod
-    def description(status):
-        return {
-            SolverStatus.Optimal: "Converged to first-order optimal solution",
-            SolverStatus.IterationLimit: "Reached iteration limit",
-            SolverStatus.TimeLimit: "Reached time limit",
-            SolverStatus.Unbounded: "Problem appears unbounded",
-            SolverStatus.LocallyInfeasible: "Local infeasibility detected",
-        }[status]
-
-    @staticmethod
-    def success(status):
-        """
-        Returns
-        -------
-        bool
-            Whether the status indicates a successful solve
-        """
-        return status == SolverStatus.Optimal
-
-
-class SolverResult:
-    """
-    The result of a solution of a :py:class:`pygradflow.problem.Problem`
-    instance with a :py:class:`pygradflow.solver.Solver`
-    """
-
-    def __init__(
-        self,
-        x: np.ndarray,
-        y: np.ndarray,
-        d: np.ndarray,
-        status: SolverStatus,
-        iterations: int,
-        num_accepted_steps: int,
-        total_time: float,
-        dist_factor: float,
-    ):
-        self._x = x
-        self._y = y
-        self._d = d
-        self._status = status
-        self.iterations = iterations
-        self.num_accepted_steps = num_accepted_steps
-        self.total_time = total_time
-        self.dist_factor = dist_factor
-
-    @property
-    def status(self) -> SolverStatus:
-        """
-        The status of the solve as a :py:class:`pygradflow.solver.SolverStatus`
-        """
-        return self._status
-
-    @property
-    def x(self) -> np.ndarray:
-        """
-        The primal solution :math:`x \\in \\mathbb{R}^{n}`
-        """
-        return self._x
-
-    @property
-    def y(self) -> np.ndarray:
-        """
-        The dual solution :math:`y \\in \\mathbb{R}^{m}`
-        """
-        return self._y
-
-    @property
-    def d(self) -> np.ndarray:
-        """
-        The dual solution :math:`d \\in \\mathbb{R}^{n}`
-        with respect to the variable bounds
-        """
-        return self._d
-
-    def __repr__(self) -> str:
-        return "SolverResult(status={0})".format(self.status)
-
-    @property
-    def success(self):
-        return SolverStatus.success(self.status)
-
-
 header_interval = 25
 
 
@@ -209,6 +81,7 @@ class Solver:
         """
         self.orig_problem = problem
         self.params = params
+        self.callbacks = Callbacks()
 
     def compute_step(
         self,
@@ -218,22 +91,9 @@ class Solver:
         display: bool,
         timer: Timer,
     ) -> StepControlResult:
-        problem = self.problem
-        params = self.params
+
         assert self.rho != -1.0
-
-        method = newton_method(problem, params, iterate, dt, self.rho)
-
-        def next_steps():
-            curr_iterate = iterate
-            while True:
-                next_step = method.step(curr_iterate)
-                yield next_step
-                curr_iterate = next_step.iterate
-
-        return controller.compute_step(
-            iterate, self.rho, dt, next_steps(), display, timer
-        )
+        return controller.compute_step(iterate, self.rho, dt, display, timer)
 
     def _deriv_check(self, x: np.ndarray, y: np.ndarray) -> None:
         from pygradflow.deriv_check import deriv_check
@@ -295,76 +155,14 @@ class Solver:
         logger.info("%20s: %45e", "Constraint violation", iterate.cons_violation)
         logger.info("%20s: %45e", "Dual violation", iterate.stat_res)
 
-    def _create_initial_iterate(
-        self, x0: Optional[np.ndarray], y0: Optional[np.ndarray]
-    ):
-        params = self.params
-        dtype = params.dtype
-        orig_problem = self.orig_problem
-        problem = self.problem
+        eval = self.evaluator
 
-        orig_lb = orig_problem.var_lb
-        orig_ub = orig_problem.var_ub
+        eval_name = Format.bold("{:>20s}".format("Evaluations"))
+        logger.info("%20s", eval_name)
 
-        if x0 is None:
-            orig_n = orig_problem.num_vars
-            x_init = np.zeros((orig_n,), dtype=dtype)
-            np.clip(x_init, orig_lb, orig_ub, out=x_init)
-        else:
-            if (x0 > orig_ub).any() or (x0 < orig_lb).any():
-                logger.warning("Initial point violates variable bounds")
-                x0 = np.clip(x0, orig_lb, orig_ub)
-
-            x_init = cast(np.ndarray, x0)
-
-        if y0 is None:
-            orig_m = orig_problem.num_cons
-            y_init = np.zeros((orig_m,), dtype=dtype)
-        else:
-            y_init = cast(np.ndarray, y0)
-
-        transform = self.transform
-        (x_init, y_init) = transform.transform_sol(x_init, y_init)
-
-        x = x_init.astype(dtype)
-        y = y_init.astype(dtype)
-
-        return Iterate(problem, params, x, y, self.evaluator)
-
-    def _create_scaling(self, x0):
-        params = self.params
-        problem = self.orig_problem
-
-        scaling_type = params.scaling_type
-
-        if params.scaling is not None:
-            assert scaling_type == ScalingType.Custom
-            return params.scaling
-
-        if scaling_type == ScalingType.NoScaling:
-            return None
-        elif scaling_type == ScalingType.GradJac:
-            obj_grad = problem.obj_grad(x0)
-            cons_jac = None
-            if problem.num_cons > 0:
-                cons_jac = problem.cons_jac(x0)
-
-            return Scaling.from_grad_jac(obj_grad, cons_jac)
-        elif scaling_type == ScalingType.Custom:
-            raise ValueError("Custom scaling requires explicit scaling")
-        else:
-            raise ValueError(f"Unknown scaling type {scaling_type}")
-
-    def _create_transformed_problem(self, x0):
-        self.scaling = self._create_scaling(x0)
-
-        orig_problem = self.orig_problem
-        params = self.params
-
-        self.transform = Transformation(orig_problem, params, self.scaling)
-        self.problem = self.transform.trans_problem
-
-        pass
+        for component, num_evals in eval.num_evals.items():
+            name = component.name()
+            logger.info("%20s: %45d", name, num_evals)
 
     def _check_terminate(self, iterate, iteration, timer):
         params = self.params
@@ -394,7 +192,9 @@ class Solver:
             return SolverStatus.Unbounded
 
     def solve(
-        self, x0: Optional[np.ndarray] = None, y0: Optional[np.ndarray] = None
+        self,
+        x0: Optional[np.ndarray] = None,
+        y0: Optional[np.ndarray] = None,
     ) -> SolverResult:
         """
         Solves the problem starting from the given primal / dual point
@@ -413,26 +213,23 @@ class Solver:
             solutions
         """
 
-        self._create_transformed_problem(x0)
+        self.transform = Transformation(self.orig_problem, self.params, x0, y0)
+
+        self.problem = self.transform.trans_problem
+
         params = self.params
         problem = self.problem
 
-        if params.validate_input:
-            self.evaluator: Evaluator = ValidatingEvaluator(self.problem, params)
-        else:
-            self.evaluator = SimpleEvaluator(self.problem, params)
+        self.evaluator = self.transform.evaluator
 
         self.penalty = penalty_strategy(self.problem, params)
         self.rho = -1.0
 
-        display = problem_display(problem, params)
+        display = solver_display(problem, params)
 
-        n = problem.num_vars
-        m = problem.num_cons
+        iterate = self.transform.initial_iterate
 
-        iterate = self._create_initial_iterate(x0, y0)
-
-        logger.info("Solving problem with {0} variables, {1} constraints".format(n, m))
+        print_problem_stats(problem, iterate)
 
         lamb = params.lamb_init
 
@@ -445,9 +242,6 @@ class Solver:
         logger.debug("Initial Aug Lag: %.10e", iterate.aug_lag(self.rho))
 
         status = None
-        start_time = time.time()
-        last_time = start_time
-        line_diff = 0
         iteration = 0
 
         logger.info(display.header)
@@ -457,18 +251,19 @@ class Solver:
         accepted_steps = 0
         iteration = 0
 
-        last_active_set = None
-        last_display_iteration = -1
-
         timer = Timer(params.time_limit)
 
+        if params.collect_path:
+            path: Optional[List[np.ndarray]] = [initial_iterate.z]
+        else:
+            path = None
+
         while True:
             status = self._check_terminate(iterate, iteration, timer)
             if status is not None:
                 break
 
-            curr_time = time.time()
-            display_iterate = curr_time - last_time >= params.display_interval
+            display_iterate = display.should_display()
 
             step_result = self.compute_step(
                 controller, iterate, 1.0 / lamb, display_iterate, timer
@@ -489,24 +284,22 @@ class Solver:
             primal_step_norm = float(np.linalg.norm(next_iterate.x - iterate.x))
             dual_step_norm = float(np.linalg.norm(next_iterate.y - iterate.y))
 
+            self.callbacks(CallbackType.ComputedStep, iterate, next_iterate, accept)
+
             if display_iterate:
-                last_time = curr_time
-                line_diff += 1
-
-                state = dict()
+                state = StateData()
                 state["iterate"] = iterate
+                state["active_set"] = lambda: step_result.active_set
+                state["obj_nonlin"] = lambda: iterate.obj_nonlin(next_iterate)
 
-                def compute_last_active_set():
-                    if last_display_iteration + 1 == iteration:
-                        return last_active_set
-                    return None
-
-                state["last_active_set"] = compute_last_active_set
-                state["curr_active_set"] = lambda: step_result.active_set
+                if problem.num_cons > 0:
+                    state["cons_nonlin"] = lambda: np.linalg.norm(
+                        iterate.cons_nonlin(next_iterate), ord=np.inf
+                    )
 
                 state["aug_lag"] = lambda: iterate.aug_lag(self.rho)
                 state["obj"] = lambda: iterate.obj()
-                state["iter"] = lambda: iteration + 1
+                state["iter"] = iteration + 1
                 state["primal_step_norm"] = lambda: primal_step_norm
                 state["dual_step_norm"] = lambda: dual_step_norm
                 state["lamb"] = lambda: lamb
@@ -514,7 +307,6 @@ class Solver:
                 state["rcond"] = lambda: step_result.rcond
 
                 logger.info(display.row(state))
-                last_display_iteration = iteration
 
             if accept:
                 # Accept
@@ -528,6 +320,9 @@ class Solver:
 
                 delta = iterate.dist(next_iterate)
 
+                if path is not None:
+                    path.append(next_iterate.z)
+
                 iterate = next_iterate
 
                 path_dist += primal_step_norm + dual_step_norm
@@ -539,8 +334,6 @@ class Solver:
                     break
 
             iteration += 1
-            last_active_set = step_result.active_set
-            # last_active_set = iterate.active_set
 
         total_time = timer.elapsed()
 
@@ -566,9 +359,16 @@ class Solver:
         y = iterate.y
         d = iterate.bounds_dual
 
-        transform = self.transform
+        (x, y, d) = self.transform.restore_sol(x, y, d)
 
-        (x, y, d) = transform.restore_sol(x, y, d)
+        result_props = dict()
+
+        if path is not None:
+            complete_path: np.ndarray = np.vstack(path).T
+            num_vars = problem.num_vars
+            result_props["path"] = complete_path
+            result_props["primal_path"] = complete_path[:num_vars, :]
+            result_props["dual_path"] = complete_path[num_vars:, :]
 
         return SolverResult(
             x,
@@ -579,4 +379,5 @@ class Solver:
             num_accepted_steps=accepted_steps,
             total_time=total_time,
             dist_factor=dist_factor,
+            **result_props,
         )
